/*
 * main.c
 *
 * projet: sudoku
 * construction d'une grille 'sudoku' de facon complementaire et/ou aleatoire
 * decembre 2005
 *
 *
 * xcfaudio AT gmail POINT com
 * http://www.xcfa.tuxfamily.org/
 *
 *
 * compilation:
 *   $ make
 * suppression des infos de debug:
 *   $ strip ./sudoku
 *
 *
 * l'idee etudiee et transformee en programme non recursif vient de 'su_doku.c' trouve sur le site:
 *   Jean-Paul Davalan Nov. 2005 <jpdvl@wanadoo.fr>
 *   http://perso.wanadoo.fr/jean-paul.davalan/jeux/logic/sudoku/su_doku.c
 *
 *
 * exemples:
 *   $ ./sudoku
 *   $ ./sudoku "32-6--5--75---286---6--7---518-----29-2---1-34-----975---8--2---852---31--4--1-58"
 *
 *
 * GNU General Public License
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * OLD ADRESS:
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * NEW ADRESS:
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/* les fichiers necessaires */
#include <stdio.h>	/* printf, fprintf, stderr, stdout, ...*/
#include <stdlib.h>	/* exit, EXIT_SUCCESS, EXIT_FAILURE, RAND_MAX, ... */
#include <time.h>	/* time */
#include <signal.h>	/* signal, SIGSEGV, ... */

/* rappel:
*  -------
*  val < 0 TRUE
*  val = 0 FALSE
*  val > 0 TRUE
*/

/* les references du programme */
#define VERSION "0.0.2"
#define NAME    "sudoku"
#define MONTH   "novembre"
#define YEAR    "2005-2009"

/* definition d'un nouveau type */
typedef enum { FALSE=0, TRUE } BOOL;

/* structure d'une grille */
typedef struct {
	int value;	/* la valeur de la case dans l'interval 1 a 9			*/
	int fixe;	/* si = 1 alors 'value' ne doit pas etre effacee		*/
	int lig;	/* 0 a 8							*/
	int col;	/* 0 a 8							*/
	int box;	/* 0 a 8							*/
	int bij [ 9 ];	/* les possibilites pour chaque case:				*/
			/*    ne doit contenir que des valeurs dans l'interval 1 a 9	*/
	int b;		/* variable de test						*/
	int v;		/* variable de test						*/
	int pos;	/* le numero de la possibilite:					*/
			/* si fixe=0 alors pos [ 0 .. max-1 ]				*/	
			/* si fixe=1 alors pos [ 1 .. max ]				*/
} ELEMENT;

/* 81 elements soit un interval de ZERO A QUATRE-VINGT */
#define MAX_ELEMENTS 81
ELEMENT elem[ MAX_ELEMENTS ];

/* pour effectuer un test avec le precedant tableau */
int copie [ MAX_ELEMENTS ];

/* contient les elements de comparaison */
int lib[ 27 ];
int m [] = {0,1,2,4,8,16,32,64,128,256};

/* l'agent comptable */
int solutions;

#define TAILLE (sizeof(elem)+sizeof(copie)+sizeof(lib)+sizeof(m)+sizeof(solutions))

void bin (char *t, int val) {
	int i,v;
	printf ("%s%2d: ",t,val);
	for(i=8;i>0;i--){
		v=1;
		v <<= i-1;
		printf ("%s ", val & v ? "1" : "0");
	}
	printf ("\n");
}

void affiche_grille (void) {
	int i;
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	printf("!------------------!\n");
	printf("! Solutions : %d\n",solutions);
	printf("!------------------!\n");	
	for(i=0;i<MAX_ELEMENTS;i++) {
		printf("%d ",  elem[ i ].value);
		if ((i+1)%27==0)	printf("\n\n");
		else if ((i+1)%9==0)	printf("\n");
		else if ((i+1)%3==0)	printf(" ");
	}
}
void init_grille (void) {
	int p,n,col,lig,box;
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	for (p=0; p<MAX_ELEMENTS; p++) {	
		copie[ p ] = 0;
		elem[ p ].value = 0;
		elem[ p ].fixe = 0;	
		lig = p / 9;
		col = p % 9;
		box = 3 * (lig/3) + (col/3);
		elem[ p ].lig = lig;
		elem[ p ].col = col;
		elem[ p ].box = box;
		for (n=0;n<9;n++) elem[ p ].bij[ n ] = 0;
		elem[ p ].b = 0;
		elem[ p ].pos = 0;
		elem[ p ].v = 0;
	}
	for (p=0; p<27; p++) lib[ p ] = 0;
	solutions = 0;
}

void make_grille (void) {
	int  u, i;
	BOOL b_avance = TRUE;
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/

	/* init du timer */
	srand(time(NULL));
	
	for (u=0; u<MAX_ELEMENTS; ) {
		if (b_avance == TRUE) {
			
			/* init 'bij[]' en aleatoire */
			for (i=0;i<9;i++) elem[ u ].bij[ i ] = i+1;
			for(i=8; i>0; i--) {
				int r,z;
				r = (int)(((i+1.0) * rand()) / (RAND_MAX + 1.0));
				z = elem[ u ].bij[ i ];
				elem[ u ].bij[ i ] = elem[ u ].bij[ r ];
				elem[ u ].bij[ r ] = z;
			}
			
			/* recup. variable de test */
			elem[ u ].v = (lib[ elem[ u ].lig ] |
				       lib[ elem[ u ].col+9 ] |
				       lib[ elem[ u ].box+18 ]);
		} else {
			if (elem[ u ].pos>=9) elem[ u ].pos = 0;
			u--;
			lib[ elem[ u ].lig ]    -= elem[ u ].b;
			lib[ elem[ u ].col+9 ]  -= elem[ u ].b;
			lib[ elem[ u ].box+18 ] -= elem[ u ].b;
		}
		
		b_avance = FALSE;
		while (elem[ u ].pos < 9) {
			int ii = elem[ u ].bij[ elem[ u ].pos ];
			elem[ u ].b = m [ ii ];
			if ((elem[ u ].v & elem[ u ].b) != elem[ u ].b) {
				lib[ elem[ u ].lig ]    |= elem[ u ].b;
				lib[ elem[ u ].col+9 ]  |= elem[ u ].b;
				lib[ elem[ u ].box+18 ] |= elem[ u ].b;
				elem[ u ].value = ii;
				elem[ u ].pos++;
				u++;
				b_avance = TRUE;
				break;
			}
			elem[ u ].pos++;
		}
	}
	solutions ++;
}

void dup (void) {
	int i;
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	for(i=0;i<MAX_ELEMENTS;i++) copie[i] = elem[i].value;
}

BOOL compare (void) {
	int i;
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	for(i=0;i<MAX_ELEMENTS;i++)
		if (copie[i] != elem[i].value) return (FALSE);
	return (TRUE);
}

void avance (void) {
	int  u = 0;
	BOOL b_avance = TRUE;
	long s = 0;
	
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	while (TRUE) {
		/* si fin atteinte, test et sortie au besoin */
		if (u == MAX_ELEMENTS) {
			if (compare () == TRUE) return;
			dup ();
			solutions ++;
			affiche_grille ();
			b_avance = FALSE;
		}

		/* on n'est jamais assez prudent */
		if (s++ > 100000) {
			printf("FIN 1 -> Plus de solution !\n\n");
			return;
		}
		
		/* ne surtout pas modifier les elements passes en ligne de commande */
		if ((b_avance == TRUE) && (elem[ u ].fixe == 1)) {
			u++;
			continue;
		}
		
		/* recup. de la variable de test */
		elem[ u ].v = (lib[ elem[ u ].lig ] |
			       lib[ elem[ u ].col+9 ] |
			       lib[ elem[ u ].box+18 ]);
		
		/* si retour en arriere */
		if (b_avance == FALSE) {
			elem[ u ].pos = 1;
			u--;
			
			/* si nous regressons de trop alors c'est fini */
			if (u<0) {
				printf("FIN 2 -> Plus de solution !\n\n");
				return;
			}
			
			/* ne surtout pas modifier les elements passes en ligne de commande */
			if (elem[ u ].fixe == 1) {
				continue;	
			
			/* modif. des variables de test */
			} else {
				lib[ elem[ u ].lig ]    -= elem[ u ].b;
				lib[ elem[ u ].col+9 ]  -= elem[ u ].b;
				lib[ elem[ u ].box+18 ] -= elem[ u ].b;		
			}
		}
		
		/* */
		b_avance = FALSE;
		while (elem[ u ].pos <= 9) {
			elem[ u ].b = m [ elem[ u ].pos ];
			if ((elem[ u ].v & elem[ u ].b) != elem[ u ].b) {
				lib[ elem[ u ].lig ]    |= elem[ u ].b;
				lib[ elem[ u ].col+9 ]  |= elem[ u ].b;
				lib[ elem[ u ].box+18 ] |= elem[ u ].b;
				elem[ u ].value = elem[ u ].pos;
				elem[ u ].pos ++;
				u ++;
				b_avance = TRUE;
				break;
			}
			elem[ u ].pos++;
		}
	}	
}

/* Decompose les options passees en ligne de commande
 */
void analyse_commandLine (char *st) {
	int   i;
	char *q;
	BOOL  bool = FALSE;	
	// int  Str [ 100 ];
	
	/*fprintf (stderr, "[%s]-> %s(%d) \n", __FILE__, __FUNCTION__, __LINE__);*/
	
	printf("\nDemande = %s\n", st);
	
	/*
	for(i=0; i<MAX_ELEMENTS; i++) 
		Str [ i ] = -1;
	
	for(i=0, q=st;*q!='\x0' && i<MAX_ELEMENTS;i++, q++) {
		int c = *q - '0';
		if(1<=c && c<=9 ) {
			Str [ i ] = c;
			bool = TRUE;
		}
	}
	for(i=0; i<MAX_ELEMENTS; i++) 
		printf("%d", Str [ i ]);
	printf("\n");
	*/
	/* recup. des elements de la grille depuis la ligne de commande */
	// for(i=0; i<MAX_ELEMENTS; i++) {
	for(i=0, q=st;*q!='\x0' && i<MAX_ELEMENTS;i++, q++) {
		int c = *q - '0';
		// if (Str [ i ] > -1) {
		if(1<=c && c<=9 ) {
			// elem[i].value = Str [ i ];
			elem[i].value = c;
			elem[i].fixe = 1;
			bool = TRUE;
		}
	}

	/* init 'lib[]' */
	for(i=0;i<27;i++) lib[ i ] = 0;
	/* si des elements valides ont ete trouves */
	for(i=0;i<MAX_ELEMENTS;i++) {
		int b;
		if(bool == TRUE) elem[ i ].pos = 1;
		if(elem[i].value != 0) {
			b = m[ elem[i].value ];
			lib[ elem[i].lig ]    += b;
			lib[ elem[i].col+9 ]  += b;
			lib[ elem[i].box+18 ] += b;
		}
	}
}

/* Signale une erreur systeme memoire et sortie propre
 */
void segfault_handler (int sig) {
	fprintf (stderr, "[%s]-> %s(line:%d) \n", __FILE__, __FUNCTION__, __LINE__);
	printf ("\n=========================================================\n");
	printf ("=  Segmentation fault                                   =\n");
	printf ("=  Adressage memoire invalide                           =\n");
	printf ("=-------------------------------------------------------=\n");
	printf ("=     You've probably found a bug in SUDOKU             =\n");
	printf ("=     Vous avez probablement trouve un bug dans SUDOKU  =\n");
	printf ("=========================================================\n\n");
	printf ("=     Announce it to me. Thank you  ;-)                 =\n");
	printf ("=     Signalez le moi. Merci  ;-)                       =\n");
	printf ("=========================================================\n\n");
	exit (EXIT_FAILURE);
}

/* Function principale
 */
int main (int argc, char *argv[]) {
	/* presentation */
	printf ("\n%s V%s -- %s %s\n\n", NAME,VERSION,MONTH,YEAR);
	
	/* place memoire occupee */
	printf ("Place memoire occupee:\n\t");
	printf ("%lu octets\n\t%lu,%lu Ko\n\n",TAILLE, TAILLE/1024, TAILLE%1024);
	
	/* if "Segmentation fault" call segfault_handler () and end of program */
	signal(SIGSEGV, segfault_handler);

	if (argc == 1) {
		/* si aucun parametre en ligne de commande */
		init_grille ();
		make_grille ();
		affiche_grille ();
	} else {
		int i;
		
		/* si au moins un parametre en ligne de commande */
		for(i=1;i<argc;i++) {
			init_grille ();
			analyse_commandLine (argv[i]);
			avance ();
		}
	}
	
	return (EXIT_SUCCESS);
}
